#include "flight_planner.h"

using airline::flight_planner;

/**
 * The constructor
 */
flight_planner::flight_planner(sc_module_name module_name, unsigned int q_size = 0) : 
  sc_module(module_name),
  in_from_master("master_to_planner"),
  out_to_master("planner_to_master"),
  clk("clock"),
  new_planning_requests(bpq_t(q_size)),
  connectivity_requests(bpq_t(q_size)),
  finished_planning_requests(bpq_t(q_size))
{
  
  SC_THREAD( receive_new_requests );
  sensitive << clk.pos();

  SC_THREAD( request_connectivity );
  sensitive << clk.pos();

  SC_THREAD( find_connectivity );
  sensitive << clk.pos();

  SC_THREAD( send_finished_requests );
  sensitive << clk.pos();

  // ... and the sets
  temp_outgoing_requests.set_size(2 * q_size);
  
  queue_size = q_size;
}
  

/**
 * The module is instantiated before the monitors are
 * instantiated. This is why we have to wait until the end of
 * elaboration to get a pointer to the monitors.
 */
void
flight_planner::before_end_of_elaboration() { 
#ifdef PROPERTY_2
  sc_module::before_end_of_elaboration();
  prop2 = sc_get_curr_simcontext()->get_observer()->get_vector_of_monitors_by_index(1);
#endif
}



/**
 * Receives requests from the master module. Adds new requests to the
 * new_planning_requests queue if there is space available, otherwise
 * blocks until space becomes available.
 */
void flight_planner::receive_new_requests() {
  dprint(4, "%s: flight_planner::%s() called\n", name(), __FUNCTION__);
  while(true) {
    dprint(4, "%s: flight_planner::%s() @ top of infinite while() loop\n", name(), __FUNCTION__);
    dprint(5, "%s: checking for new requests on %s\n", name(), in_from_master.name());
      
    tlm_tag_t t;
    if (! in_from_master->nb_can_get( &t )) {
      dprint(6, "%s: no tokens available on %s\n", name(), in_from_master.name());
      dprint(5, "%s: suspending %s() until new tokens become available on %s\n", 
	     name(), __FUNCTION__, in_from_master.name());
      
      tlm_tag_t tag;
      wait(in_from_master->ok_to_get(&tag));
      dprint(5, "%s: %s() is awake. New tokens have become available on %s\n", 
	     name(), __FUNCTION__, in_from_master.name());
    }
    
    token_t* req = new token_t();
    in_from_master->nb_get(req);
    dprint(5, "%s: consumed request_id %d with status %s from %s\n", 
	   name(),
	   req->get_request()->get_request_id(), 
	   req->get_payload() == PLAN ? "PLAN" : "OTHER",
	   in_from_master.name());
    
#ifdef PROPERTY_2
    for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
      ((property2*) (*it))->flight_planner_receive_request_from_master(req->get_request()->get_request_id());
    }
#endif
    
      
    if (req->get_payload() == PLAN) {
      unsigned int curr_size = new_planning_requests.size();
      dprint(6, "%s: queue new_planning_requests currently has %d items\n", name(), curr_size);
      
      if (curr_size >= queue_size) {
	dprint(6, "%s: new_planning_requests does not have space for a new token.\n", name());
	dprint(5, "%s: suspending get_new_requests() until new_planning_requests has space\n", name());
	wait(new_requests_nonfull);
	dprint(5, "%s: get_new_requests() is awake. new_planning_requests is not full\n", name());
      }
      
      // Either there is space, or we have been awakened because there is space
      assert(new_planning_requests.size() < queue_size);
      
      dprint(6, "%s: new_planning_requests has space for a new token.\n", name());
      dprint(5, "%s: adding request %d to internal queue \"new_planning_requests\"\n", 
	     name(), 
	     req->get_request()->get_request_id());
      
      new_planning_requests.push(req->get_request());
      new_requests_nonempty.notify(SC_ZERO_TIME);
      
      dprint(4, "%s: suspending get_new_requests() until the next clock tick\n", name());
      wait();
    } // it was a new request
    
    else {
      handle_special_request(req->get_request(), req->get_payload());
    }
  } // receiving loop
} //receive_new_requests()



/** 
 * For each pending request in the incoming queue finds connectivity
 * information (that is, legs between the requested cities).
 */
void flight_planner::request_connectivity() {
  dprint(4, "%s: flight_planner::%s() called\n", name(), __FUNCTION__);
  while(true) {
    dprint(4, "%s: flight_planner::%s() @ top of connectivity loop \n", name(), __FUNCTION__);
    unsigned int curr_size = new_planning_requests.size();
    dprint(5, "%s: queue new_planning_requests currently has %d items\n", name(), curr_size);
    
    if (curr_size == 0) {
      dprint(6, "%s: pending requests queue is empty\n", name());
      dprint(5, "%s: Suspending %s until there is something in the queue\n", name(), __FUNCTION__);
      wait(new_requests_nonempty);
      dprint(5, "%s: request_connectivity(): new_requests is no longer empty.\n", name());
    }

    assert(new_planning_requests.size() > 0);

    TRANS_TYPE curr_request = new_planning_requests.toppop();
    new_requests_nonfull.notify(SC_ZERO_TIME);
    
    dprint(5, "%s: dequeued request %d\n", name(), curr_request->get_request_id());
    if (curr_request->is_return()) {
      dprint(6, "%s: this is a return trip\n", name());
      
      // Pushing two separate requests, one for each direction.
      request* return_flight = new request();
      *return_flight = *curr_request;
      
      // Indicate that these two requests are a part of a set.
      curr_request->set_req_sequence(1);
      return_flight->set_req_sequence(2);
      
      // The outbound flight is given higher priority
      curr_request->update_priority( +1);

      // The return flight is defined
      return_flight->set_date(curr_request->get_date2());
      return_flight->set_source(curr_request->get_destination());
      return_flight->set_destination(curr_request->get_source());
      
      CONDPRINT(6, curr_request->print_me());
      CONDPRINT(6, return_flight->print_me());
      
      unsigned int conn_size = connectivity_requests.size();
      dprint(5, "%s: Connectivity queue has %d items\n", name(), conn_size);
      
      
      // Adding the outbound flight connectivity request...
      b_add_request_to_conn_queue(curr_request);     

      // We have a race condition. Need to resolve the signal before
      // we add the other request.
      wait();
      
      // Adding the inbound flight request
      b_add_request_to_conn_queue(return_flight);
      

    } // Return trips

    else {
      dprint(6, "%s: this is a one-way trip\n", name());
      b_add_request_to_conn_queue(curr_request);
    } // One-way trips

    dprint(5, "%s: Done requesting connectivity for request %d\n", name(), curr_request->get_request_id());
    dprint(4, "%s: Suspending %s until next clock tick\n", name(), __FUNCTION__);
    wait();
  } // while
} // request_connectivity()


/**
 * Adds a request to the connectivity queue (conn_queue). If the queue
 * is full, blocks until the queue has some space (waits for
 * notification of conn_queue_nonempty).
 */
void flight_planner::b_add_request_to_conn_queue(TRANS_TYPE req) {
  dprint(4, "%s: flight_planner::%s() called\n", name(), __FUNCTION__);
  assert (connectivity_requests.size() <= queue_size);
  if (connectivity_requests.size() == queue_size) {
    dprint(6, "%s: conn_queue full. Suspending %s until the queue has some space\n", name(), __FUNCTION__);
    wait(conn_queue_nonfull);
    dprint(6, "%s: %s() is awake. conn_queue is no longer full.\n", name(), __FUNCTION__);
  }
  
  dprint(9, "%s: before assert, the size of the conn_requests queue is %d\n", 
	 name(), connectivity_requests.size());
  assert(connectivity_requests.size() < queue_size);

  unsigned int old_size = connectivity_requests.size();
  
  dprint(5, "%s: Adding request %d (%d) to the conn queue\n", 
	 name(), req->get_request_id(), req->get_req_sequence());
  connectivity_requests.push(req);
  
  assert(connectivity_requests.size() == old_size + 1);
  
  conn_queue_nonempty.notify(SC_ZERO_TIME);
}



/**
 * Finds connecting flights between two points. Essentially a simple
 * database query with an increasing number of connecting flights, up
 * to some limit.
 */
void flight_planner::find_connectivity() {
  dprint(4, "%s: flight_planner::%s() called\n", name(), __FUNCTION__);
  while(true) {
    dprint(4, "%s: flight_planner::%s(): start new connectivity loop \n", name(), __FUNCTION__);
    dprint(6, "%s: checking for pending requests on the conn_requests queue\n", name());
    unsigned int curr_size = connectivity_requests.size();

    dprint(5, "%s: %d items currently in the queue\n", name(), curr_size);

    if (curr_size == 0) {
      dprint(5, "%s: suspending %s until new requests are added to the conn_requests\n", name(), __FUNCTION__);
      wait(conn_queue_nonempty);
    }
    
    assert(connectivity_requests.size() > 0);
    unsigned int old_size = connectivity_requests.size();
    
    TRANS_TYPE curr_request = connectivity_requests.toppop();

    assert(connectivity_requests.size() == old_size - 1);

    dprint(5, "%s: consumed request %d (%d) from the conn_requests queue.\n",
	   name(), curr_request->get_request_id(), curr_request->get_req_sequence());
    conn_queue_nonfull.notify(SC_ZERO_TIME);

    dprint(9, "%s: after conn_queue_nonfull.notify(), the size of conn_requests is %d\n", 
	   name(), connectivity_requests.size());

    // do some work to find connectivity
    // in this version the connectivity is randomly generated
    list_t* conn_list = new list_t();
    int num_legs = 1 + ( rand() % 3 );
    int old_destination = REQUEST_UNDEF_VAL;

    for (int i=0; i<num_legs; i++) {
      int source = REQUEST_UNDEF_VAL;
      int destination = REQUEST_UNDEF_VAL;
      if (i == 0) 
	source = curr_request->get_source();
      else 
	source = old_destination;
      
      if (i == num_legs - 1)
	destination = curr_request->get_destination();
      else
	destination = rand() % 20;
      
      pair_t* p = new pair_t(source, destination);
      old_destination = destination;
      conn_list->push_back(p);
    }
    
    dprint(5, "%s: updating the request with the flight legs\n", name());
    curr_request->set_connecting_flights(conn_list);
    dprint(6, "%s: added connectivity data to request %d (%d)\n", 
	   name(), curr_request->get_request_id(), curr_request->get_req_sequence());

    CONDPRINT(6, curr_request->print_me());

    // Push the request to the outgoing queue...
    unsigned int fin_size = finished_planning_requests.size();
    dprint(9, "%s: finished requests queue has %d members\n", name(), fin_size);
    
    if (fin_size == queue_size) {
      dprint(9, "%s: suspending %s until finished_req_queue has space for tokens\n", name(), __FUNCTION__);
      wait(finished_requests_nonfull);      
    }

    assert(finished_planning_requests.size() < queue_size);

    dprint(9, "%s: finished_req_queue has space for tokens. Adding %d (%d)\n", 
	   name(), curr_request->get_request_id(), curr_request->get_req_sequence());
    finished_planning_requests.push(curr_request);
    finished_requests_nonempty.notify(SC_ZERO_TIME);
    
    dprint(9, "%s: suspending %s until the next clock tick\n", name(), __FUNCTION__);
    wait();
  }
}




/**
 * Collates and sends finished requests back to the master module. In
 * cases of multiple requests (that is, those with a return flight),
 * it waits to collect both requests before sending them. However, for
 * efficiency, it makes a guess about the missing token and as soon as
 * the second token arrives, it sends a correction request.
 */
void flight_planner::send_finished_requests() {
  dprint(4, "%s: flight_planner::%s() called\n", name(), __FUNCTION__);

  while(true) {
    dprint(4, "%s: flight_planner::%s() begin a new sending loop\n", name(), __FUNCTION__);

    unsigned int fin_size = finished_planning_requests.size();
    dprint(5, "%s: queue finished_planning_requests has %d elements\n", name(), fin_size);
    
    if (fin_size == 0) {
      dprint(6, "%s: suspending %s until finished_planning_requests is non-empty\n", name(), __FUNCTION__);
      wait(finished_requests_nonempty);
    }

    dprint(5, "%s: send_finished_requests(): finished_planning_requests is non-empty\n", name());
    assert (finished_planning_requests.size() > 0 );

    TRANS_TYPE curr_req = finished_planning_requests.toppop();
    finished_requests_nonfull.notify(SC_ZERO_TIME);
    
    dprint(6, "%s: dequeued transaction %d (%d)\n",
	   name(), curr_req->get_request_id(), curr_req->get_req_sequence());
    
    if (curr_req->is_return()) {
      dprint(6, "%s: this is a return flight request.\n", name());
      
      std::pair<bis_t::iterator, bool> p = temp_outgoing_requests.insert(curr_req);

      dprint(7, "%s: Contents of temp_outgoing_requests follow\n", name());
      CONDPRINT(7, temp_outgoing_requests.print_me());
      
      if ( p.second ) {
	dprint(6, "%s: This is the first component of this 2-part request\n", name());
	dprint(6, "%s: Guessing the connectivity of the in-bound flight\n", name());
	
	curr_req->set_connecting_flights2( request::reverse_connecting_flights( curr_req->get_connecting_flights() ) );
	curr_req->set_speculative_bit_to(true);
	CONDPRINT(6, curr_req->print_me());
	
	tlm_tag_t tag;
	if (! out_to_master->nb_can_put( &tag )) {
	  dprint(6, "%s: Suspending %s until the channel becomes available.\n", name(),__FUNCTION__);
	  wait(out_to_master->ok_to_put(&tag));
	  dprint(6, "%s: send_finished_requests() has been awoken by a signal,\n", name());
	}
	dprint(5, "%s: sending request %d up the channel\n", name(), curr_req->get_request_id());


	// Create a new token for this request
	token_t* new_token = new token_t(curr_req, PLANDONE);
	out_to_master->nb_put(new_token);
	
#ifdef PROPERTY_2
	for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
	  ((property2*) (*it))->flight_planner_send_request_to_master(curr_req->get_request_id());
    }
#endif

	dprint(4, "%s: suspending %s until the next clock tick.\n", name(), __FUNCTION__);
	wait();
      } // Speculative execution

      else { // Start correction phase
	dprint(5, "%s: This is the second component of this 2-part request\n", name());
	dprint(5, "%s: The original component: \n", name());
	CONDPRINT(5, (*(p.first))->print_me());
	
	dprint(5, "%s: The current component: \n", name());
	CONDPRINT(5, curr_req->print_me());

	dprint(5, "%s: Creating a correction request with the correct information\n", name());
	request* copy = new request();
	*copy = **(p.first);
	
	copy->set_connecting_flights2(curr_req->get_connecting_flights());
	
	dprint(5, "%s: A printout of the correction request follows: \n", name());
	CONDPRINT(5, copy->print_me());

	temp_outgoing_requests.erase(p.first);

	dprint(6, "%s: The set of partially completed requests: \n", name());
	CONDPRINT(6, temp_outgoing_requests.print_me());

	// Send the correction request up the channel
	tlm_tag_t tag;
	if (! out_to_master->nb_can_put( &tag )) {
	  dprint(6, "%s: Suspending %s until the channel becomes available.\n", name(), __FUNCTION__);
	  wait(out_to_master->ok_to_put(&tag));
	  dprint(6, "%s: send_finished_requests() has been awoken by a signal,\n", name());
	}
	dprint(5, "%s: sending UPDATED request %d\n", name(), curr_req->get_request_id());

	token_t* new_token = new token_t(curr_req, UPDATED);
	out_to_master->nb_put(new_token);
#ifdef PROPERTY_2
	for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
	  ((property2*) (*it))->flight_planner_send_request_to_master(curr_req->get_request_id());
	}
#endif
	
	dprint(4, "%s: suspending %s until the next clock tick.\n", name(), __FUNCTION__);
	wait();
      } // Correction phase
    } // this is a return trip
      
    else { // 1-way trip
      dprint(5, "%s: This is a 1-way trip\n", name());
      
      tlm_tag_t tag;
      if ( !(out_to_master->nb_can_put( &tag ))) {
	// The channel is currently full. Waiting until it is free again...
	dprint(6, "%s: channel on %s has no space for a token\n",  name(), out_to_master.name());
	dprint(6, "%s: Suspending %s until the channel becomes available.\n", name(), __FUNCTION__);
	wait(out_to_master->ok_to_put(&tag));
	dprint(6, "%s: send_finished_requests() has been awoken by a signal,\n", name());
      }

      assert (out_to_master->nb_can_put( &tag ));
	
      dprint(5, "%s: sending request %d\n", name(), curr_req->get_request_id());
      
      // Create a new token for this request
      token_t* new_token = new token_t(curr_req, PLANDONE);
      out_to_master->nb_put(new_token);
#ifdef PROPERTY_2
      for(vect_mon_t::const_iterator it = prop2->begin(); it != prop2->end(); it++) {
	((property2*) (*it))->flight_planner_send_request_to_master(curr_req->get_request_id());
      }
#endif
      
      dprint(4, "%s: suspending %s until the next clock tick.\n", name(), __FUNCTION__);
      wait();
    } // This is a one-way trip
  } // sending loop
}



/**
 * Handles special requests from the master and/or the DB agent. 
 */
void flight_planner::handle_special_request(TRANS_TYPE t, PAYLD_TYPE p) {
  dprint(9, "%s: flight_planner::%s() called\n", name(), __FUNCTION__);
  dprint(1, "%s: handle_special_request() not implemented yet. \n", name());
  exit(1);
}
